/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*

	bfgs.c

		Implementation of Broyden-Fletcher-Goldfarb-Shanno
	optimization algorithm, as a first step to evaluate
	Broyden family optimization algorithms, of which the
	BFGS method is also a member.

	5/18/87 - modified to take advantage of symmetry in
		update_hessian

	1/14/88 - changed termination condition to ACCEPTABLE
		constant - 0.01 for speech experiments; corresponds
		to MEAN squared error, an average error of 0.1 per
		sample.

	1/20/88 - modified to avoid redundant gradient computation

	2/1/88 - modified for saving networks at different values of
		the objective function

	3/10/88 - changed filename to reflect actual LMS value

*/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include "vector.h"

extern int initialize();
extern int finalize();
extern double objective_function();
extern int compute_gradient();
extern double find_min_along();
extern double dot_product();

extern int space;

#define EPSILON 0.000000000001

#define ACCEPTABLE 0.00000001

#ifdef TERMIN
#define NUM_GOALS 20

#define BASE_GOAL 0.010

double goal[NUM_GOALS];
int gndx;
#endif


char corpus[100];
int iterations;
int function_evaluations;
int gradient_evaluations;




struct vector *neg_vect(a)
struct vector *a;
{
static struct vector negative;
int index;

for (index=0; index<space; index++)
	negative.val[index] = -(a->val[index]);
return(&negative);
} 




int vector_sum(a,b,c)
/* a gets vector b + c */
struct vector *a, *b, *c;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = b->val[index] + c->val[index];
} 



int vector_difference(a,b,c)
struct vector *a, *b, *c;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = b->val[index] - c->val[index];
} 



int vector_scale(a,b,s)
struct vector *a, *b;
double s;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = b->val[index] * s;
} 


make_identity(matrix)
struct square *matrix;
{
/*      Initialize matrix to identity matrix    */

int i,j;

for (i=0; i<space; i++)
	for (j=0; j<space; j++)
		if (i == j) matrix->val[i][j] = 1.000;
		else matrix->val[i][j] = 0.000;
}




int product(a,m1,v)
struct vector *a,*v;
struct square *m1;
/*
Computes product of square matrix and column vector as column vector
*/
{
double total;
int i,j;

for (i=0; i<space; i++)
	{
	total = 0;
	for (j=0; j<space; j++)
		total += m1->val[i][j] * v->val[j];
	a->val[i] = total;
	}
}





int update_hessian(h,x,s,g,alpha,gnew)
struct square *h;
struct vector *x, *s, *g, *gnew;
double alpha;
{
double delgam_scale,gscale,bfgs_scale, value;
struct vector gamma, delta;
struct vector hg;
int i,j;

/*
Side effect - updates x
*/
vector_scale(&delta,s,alpha);
vector_sum(x,x,&delta);
vector_difference(&gamma,gnew,g);
delgam_scale = dot_product(&delta,&gamma);
product(&hg,h,&gamma);
gscale = dot_product(&gamma,&hg);
bfgs_scale = (1 + gscale/delgam_scale)/delgam_scale;
for (i=0; i<space; i++)
	for (j=i; j<space; j++)
		{
		value = ((delta.val[i] * delta.val[j] * bfgs_scale)
			- (delta.val[i] * hg.val[j] +
			hg.val[i] * delta.val[j]) / delgam_scale);
		h->val[i][j] += value;
		if (i != j) h->val[j][i] += value;
		}
iterations++;
}




main(argc,argv)
int argc;
char *argv[];
{
static struct vector x0;
struct vector grad1,grad2,*search_vect,v1;
struct vector *gradient,*new_grad, *temp;
static struct square hessian;
double alpha,objective_value,min;
int done,strikes,i;
char net_name[100];


space = initialize(argc,argv,&x0);

finalize("net.testio.fin",&x0);

make_identity(&hessian);
iterations = 0;
done = 0;	strikes = 0;
min = ACCEPTABLE;

objective_value = objective_function(&x0);
printf("%d\tObjective value\t%.10f\n",iterations,objective_value); 
gradient = &grad1;
new_grad = &grad2;

#ifdef TERMIN
for (gndx=0; gndx<NUM_GOALS; gndx++)
	goal[gndx] = BASE_GOAL * (1.00 - ((double) gndx)/NUM_GOALS);

if (objective_value < goal[NUM_GOALS-1]) done = 1;
else compute_gradient(&x0,gradient);

gndx = 0;
#else
if (objective_value < min) done = 1;
else compute_gradient(&x0,gradient);
#endif

while (!done)
	{
#ifdef DEBUG
	for (i=0; i<space; i++)
		printf("%.10f\t",x0.val[i]);
	printf("\n");
#endif
#ifdef TERMIN
	while ((gndx < NUM_GOALS) && (objective_value <goal[gndx]))
		{
		printf("Goal Index %d\t",gndx);
		printf("Met Goal %f:\t%d\t%d\t%d\n",goal[gndx],iterations,
			function_evaluations,gradient_evaluations);
		sprintf(net_name,"net.t%f.fin",goal[gndx]);
		finalize(net_name,&x0);
		gndx++;
		}
	if (gndx == NUM_GOALS) done = 1;
#else
	if (objective_value < min) done = 1;
#endif
	else
		{
		product(&v1,&hessian,gradient);
		search_vect = neg_vect(&v1);
		alpha = find_min_along(&objective_value,&x0,
			gradient,search_vect,objective_function,
			compute_gradient,new_grad);
#ifndef SILENT
		printf("%d\t%d\t%d\tObjective value\t%.10f\t",iterations,
			function_evaluations,
			gradient_evaluations,
			objective_value); 
		printf("Alpha %.10f\n",alpha);
		fflush(stdout);
#endif
		if (alpha <= EPSILON)
			{
			strikes++;
			make_identity(&hessian);
			}
		else
			{
			strikes = 0;
			update_hessian(&hessian,&x0,search_vect,
			gradient,alpha,new_grad);
/*
SWITCH GRADIENTS
*/
			temp = gradient;
			gradient = new_grad;
			new_grad = temp;
			}
		finalize("net.tmp.tmp",&x0);
		if (strikes >2) done = 1;
		}
	}
printf("%d\t%d\t%d\tObjective value\t%.10f\n",iterations,
	function_evaluations,
	gradient_evaluations,
	objective_value); 
if (objective_value < min)
printf("Iterations to solve = %d\n",iterations);
else printf("Iterations to failure = %d\n",iterations);
finalize("net.x.fin",&x0);
}


