/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*

	dfp.c

		Implementation of Davidon-Fletcher-Powell
	optimization algorithm, as a first step to evaluate
	Broyden family optimization algorithms, of which the
	BFGS method is also a member.

	1/14/88 - changed termination condition to ACCEPTABLE
		constant - 0.01 for speech experiments; corresponds
		to MEAN squared error, an average error of 0.1 per
		sample.

	1/20/88 - modified for changes to avoid extra gradient comp

*/


#include <math.h>
#include <stdio.h>
#include <string.h>
#include "vector.h"

extern int initialize();
extern int finalize();
extern double objective_function();
extern int compute_gradient();
extern double find_min_along();
extern double dot_product();

extern int space;

#define EPSILON 0.00000001

#define ACCEPTABLE 0.001

char corpus[100];
int iterations;
int function_evaluations;
int gradient_evaluations;


struct vector *neg_vect(a)
struct vector *a;
{
static struct vector negative;
int index;

for (index=0; index<space; index++)
	negative.val[index] = -(a->val[index]);
return(&negative);
} 


int vector_difference(a,b,c)
struct vector *a, *b, *c;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = b->val[index] - c->val[index];
} 




int vector_sum(a,b,c)
/* a gets vector b + c */
struct vector *a, *b, *c;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = b->val[index] + c->val[index];
} 




int vector_scale(a,b,s)
struct vector *a, *b;
double s;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = b->val[index] * s;
} 


make_identity(matrix)
struct square *matrix;
{
/*	Initialize matrix to identity matrix	*/

int i,j;

for (i=0; i<space; i++)
	for (j=0; j<space; j++)
		if (i == j) matrix->val[i][j] = 1.000;
		else matrix->val[i][j] = 0.000;
}




int product(a,m1,v)
struct vector *a,*v;
struct square *m1;
/*
Computes product of square matrix and column vector as column vector
*/
{
double total;
int i,j;

for (i=0; i<space; i++)
	{
	total = 0;
	for (j=0; j<space; j++)
		total += m1->val[i][j] * v->val[j];
	a->val[i] = total;
	}
}



mult(a,v,m1)
struct vector *a,*v;
struct square *m1;
/*
Forms product of row vector and square matrix as column vector
*/
{
double total;
int i,j;

for (i=0; i<space; i++)
	{
	total = 0;
	for (j=0; j<space; j++)
		total += m1->val[j][i] * v->val[j];
	a->val[i] = total;
	}
}







int update_hessian(h,x,s,g,alpha,gnew)
struct square *h;
struct vector *x, *s, *g, *gnew;
double alpha;
{
double delgam_scale,gscale;
struct vector gamma, delta;
struct vector hg,gh;
int i,j;

vector_scale(&delta,s,alpha);
vector_sum(x,x,&delta);
/* gnew = compute_gradient(x); */
vector_difference(&gamma,gnew,g);
delgam_scale = dot_product(&delta,&gamma);
product(&hg,h,&gamma);
gscale = dot_product(&gamma,&hg);
mult(&gh,&gamma,h);
for (i=0; i<space; i++)
	for (j=0; j<space; j++)
		h->val[i][j] += ((delta.val[i] * delta.val[j])/delgam_scale -
			(hg.val[i] * gh.val[j]) / gscale);
/* vector_copy(gnew,g); */
iterations++;
}






main(argc,argv)
int argc;
char *argv[];
{
static struct vector x0;
struct vector grad1,grad2,*search_vect,v1;
struct vector *gradient,*new_grad, *temp;
static struct square hessian;
double alpha,objective_value,min;
int done,strikes;

space = initialize(argc,argv,&x0);

make_identity(&hessian);
iterations = 0;
done = 0;	strikes = 0;
min = ACCEPTABLE;

gradient = &grad1;
new_grad = &grad2;

objective_value = objective_function(&x0);
printf("%d\tObjective value\t%f\n",iterations,objective_value); 
compute_gradient(&x0,gradient);
while (!done)
	{
	if (objective_value < min) done = 1;
	else
		{
		product(&v1,&hessian,gradient);
		search_vect = neg_vect(&v1);
		alpha = find_min_along(&objective_value,&x0,
			gradient,search_vect,objective_function,
			compute_gradient,new_grad);
		printf("%d\t%d\t%d\tObjective value\t%f\t",iterations,
			function_evaluations,
			gradient_evaluations,
			objective_value); 
		printf("Alpha %.10f\n",alpha);
		fflush(stdout);
                if (alpha <= EPSILON)
			{
			strikes++;
			make_identity(&hessian);
			}
                else
			{
			strikes = 0;
			update_hessian(&hessian,&x0,search_vect,
                        gradient,alpha,new_grad);
			temp = gradient;
			gradient = new_grad;
			new_grad = temp;
			}
		finalize("net.dfp.tmp",&x0);
		if (strikes >2) done = 1;
		}
	}
if (objective_value < min)
printf("Iterations to solve = %d\n",iterations);
else printf("Iterations to failure = %d\n",iterations);
finalize("net.dfp.fin",&x0);
}

