/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*

	Line_search.c

		This is an auxiliary module for
	executing line search as a part of optimization.


	The necessary variables are the search direction,
	and an evaluation function.

	The source for this algorithm is R. Fletcher,
	Practical Methods for Optimization, John Wiley, 1980
	pp 25ff.


	In order to be fairly general, the following functions
	are required: the error function to be minimized, and
	gradient. Given is the initial function value, gradient
	vector and initial alpha. Returned is the minimizing
	value for alpha.


	9/23/87 - extensively modified based on further insight
	into Fletcher's descriptions, including errors.

	A more useful reference being Dennis & Schnabel, Numerical
	Methods for Unconstrained Optimization and Nonlinear Equations,
	Prentice-Hall, 1983.


	12/5/87 - modified for vcc, function name args

	12/15/87 - modified for VERBOSE mode switch

	1/20/88 - modified for new gradient computation; leaves new
		gradient value at new_grad

	4/4/89 - corrected convergence problem

	4/6/89 - added modifications from Fletcher 2nd edition

*/

#include "vector.h"
#include <stdio.h>
#include <math.h>

#ifndef FLETCHER_1ST
#define FLETCHER_2ND
#endif

#ifdef FLETCHER_2ND
#define MONOTONIC
#define WOLFE_TEST
#define EXTRAPOLATE_INTERVAL
#define NEW_MAX_ALPHA
#endif

#ifndef FIRST_ORDER
#ifndef QUASI_NEWTON
#define QUASI_NEWTON
#endif
#endif

/*
	Line Search Constraints

	RHO in (0,1/2)

	SIGMA in (RHO,1)	0.1 ==> strict line search
				0.9 ==> weak line search

	TAU in (0,SIGMA)

*/

#define RHO 0.1
#ifndef SIGMA
#define SIGMA 0.5
#endif
#define TAU 0.05
#define MAX_EXTRAPOL 9

/*	Termination constraint		*/

#define EPSILON		0.000000000000000001
#define EPSILON10	0.000000000000000010

#define F_MIN 0.000

#ifndef EXTENSOR
#define EXTENSOR 2.00
#endif

int space;			/* global space size */


double dot_product(a,b)
struct vector *a, *b;
{
int index;
double total;

total = 0;
for (index=0; index<space; index++)
	total += a->val[index] * b->val[index];
return(total);
}




int extend(new,a,b,s)
struct vector *new, *a, *b;
double s;
{
int index;

for (index=0; index<space; index++)
	new->val[index] = a->val[index] + s * b->val[index];
} 


int quadratic(a1,f1,f1_prime,a2,f2,res)
double a1,f1,f1_prime,a2,f2,*res;
/*
quadratic match given values (a1,f1) and (a2,f2) and
slope f1_prime at a1. Stores new point at res;
returns flag=1 if minimum found.
*/

{
double delta,dx;

delta = a2 - a1;
dx = delta * f1_prime;

if ((f2 -f1) <= dx)
	{
#ifdef VERBOSE
	printf("\t\t\tQuadratic fails: use default");
#endif
	return(0);
	}
*res = a1 + 0.5 * delta /(1 + (f1 - f2)/(dx));
#ifdef VERBOSE
		printf("\t\t\tQuadratic March: %.10f\n",*res);
#endif
return(1);
}


int cubic(a1,f1,f1_prime,a2,f2,f2_prime,res)
double a1,f1,f1_prime,a2,f2,f2_prime,*res;
/*
Cubic match given points (a1,f1) and (a2,f2) and
slopes f1_prime at a1, f2_prime at a2. New point written at res.
Returns ok=1 for found minimum, 0 for no minimum.
*/
{
double delta,lambda,new,w,z,b;
double rad;
int ok;

lambda = (a2 - a1);
delta = (f2 - f1) / lambda;
b = f1_prime + f2_prime - 2 * delta;

ok = 1;
if (b == 0)
	{
/*	Quadratic form: if f1_prime == f2_prime, (straight line) -
	Sign conditions for min/max dictate that for minimum,
	f2_prime > f1_prime.
*/
#ifdef VERBOSE
		printf("\t\t\tDegenerate Cubic: ");
#endif
	if (f2_prime <= f1_prime)
		{
		ok = 0;
#ifdef VERBOSE
		printf("Quadratic match failed - use default\n");
#endif
		}
	else
		{
		new = (a1 * f2_prime - a2 * f1_prime) / (f2_prime - f1_prime);
#ifdef VERBOSE
		printf("Quadratic match - new alpha: %.10f\n",new);
#endif
		}
	}
else 
	{
	z = b - delta;
	rad = z * z - f1_prime * f2_prime;
/*
	If rad is 0, inflection is saddle point - extend by default action.
	If rad is negative, imaginary minima - also extend by default.
*/

	if (rad <=0)
		{
		ok = 0;
#ifdef VERBOSE
		printf("\t\t\tSaddle Point or Imaginary min: rad = %f\n",rad);
#endif
		}
/*
	It turns out that the sign conditions for the inflection points
	guarantee that the negative choice for w always yields the
	minimum rather than the maximum. Why is this?
*/
	else
    		{
		w = sqrt(rad);
		new = a1 + lambda * (1.00 -(f2_prime - w + z)/(3 * b));
#ifdef VERBOSE
		printf("\t\t\tCubic Match to new alpha: %.10f\n",new);
#endif
		}
	}
*res = new;
return(ok);
}



int extrapolate(a1,f1,f1_prime,a2,f2,f2_prime,amax,res,bracket)
double a1,f1,f1_prime,a2,f2,f2_prime,amax,*res;
int bracket;
/*
Extrapolation given points (a1,f1) and (a2,f2) and
slopes f1_prime at a1, f2_prime at a2, and amax.
*/
{
double lambda,new;
double low, high, limit;
int success;

/*
	Check extrapolation preconditions
*/

#ifdef VERBOSE
printf("\t\tExtrapolate from <%.10f, %.10f> less than %.10f\n",a1,a2,amax);
#endif

lambda = (a2 - a1);
if (lambda <= EPSILON)
	{
	if ((amax - a2) <= EPSILON10)
		{
#ifdef VERBOSE
		printf("\t\tExtrapolation failed - search space vanished\n");
#endif
		return(0);
		}
	else
		{
		new = a2 + TAU * (amax - a2);
		*res = new;
#ifdef VERBOSE
		printf("\t\tExtrapolation interval collapsed - try new guess %.10f\n",new);
#endif
		return(1);
		}
	}

/*
	Get extrapolation value
*/

success = cubic(a1,f1,f1_prime,a2,f2,f2_prime,&new);
if (new < 0 ) success = 0;
/*
How the extrapolation failure is treated turns out to have an
effect on average convergence counts for Rosenbrock's function.
The more conservative a2+1.3 *lambda works pretty well.
*/
if (success == 0) 
	{
	if (bracket) new = amax;
	else new = a2 +  EXTENSOR * lambda;
#ifdef VERBOSE
printf("\t\t\tExtrap failed: guess (brkt = %d) : %.10f\n",bracket,new);
#endif
	}
/*
check constraints on alpha new
*/

if (bracket)
	{
	lambda = amax - a2;
	low = a2 + TAU * lambda;
	high = amax - TAU * lambda;
#ifdef VERBOSE
printf("\t\t\tBracket Extrap limits: low: %.10f,high: %.10f\n",low,high);
#endif
	}
else
	{
	high = a2 + lambda * MAX_EXTRAPOL;
	if (amax < high) high = amax;
	low = a2 + TAU * lambda;
	if (high < low) low = high;
#ifdef VERBOSE
printf("\t\t\tNo Bracket Extrap limits: low: %.10f,high: %.10f\n",low,high);
#endif
	}
if (new < low)
	{
#ifdef VERBOSE
	printf("\t\tExtrapolant (%.10f) too small: clamp to %.10f\n",new,low);
#endif
	new = low;
	}
else if (new > high)
  	{
#ifdef VERBOSE
	printf("\t\tExtrapolant (%.10f) too large: clamp to %.10f\n",new,high);
#endif
	new = high;
	}
#ifdef VERBOSE
else printf("\t\tExtrapolant (%.10f) OK\n",new);
#endif
*res = new;
return(1);
}


/*
Interpolation:
	inputs -
	(a1,f1) and slope f1_prime at a1.
	(a2,f2) and slope f2_prime at a2 if flag = 1
*/

int interpolate(a1,f1,f1_prime,a2,f2,f2_prime,f2_flag,res)
double a1,f1,f1_prime,a2,f2,f2_prime,*res;
int f2_flag;
{
double delta,new;
double low, high, range;
int success;

/*
Check Interpolation pre-conditions
*/

#ifdef VERBOSE
printf("\t\tInterpolate between %.10f and %.10f\n",a1,a2);
#endif

delta = (a2 - a1);
if (-delta * f1_prime <= EPSILON)
	{
#ifdef VERBOSE
	printf("\t\tInterpolate limits failed : %.10f\n",-delta * f1_prime);
#endif
	return(0);
	}
/*
	Now, interpolate point by quadratic or cubic methods
	based on f2_flag
*/

if (f2_flag == 1)
	{
	success = cubic(a1,f1,f1_prime,a2,f2,f2_prime,&new);
	}
else
	{
	success = quadratic(a1,f1,f1_prime,a2,f2,&new);
	}

/*
Failed interpolation - go to center of interval
*/
if (success == 0)
	{
	new = a1 + 0.5 * delta;
#ifdef VERBOSE
	printf("\tFailed interpolation: guess %.10f\n",new);
#endif
	}
/*
	check interpolate post-conditions
	new value is within (a1,a2) and not with TAU*range of ends
*/
range = TAU * delta;
low = a1 + range;
high = a2 - range;

if (new < low)
	{
#ifdef VERBOSE
	printf("\t\tInterpolant (%.10f) too small: clamp to %.10f\n",new,low);
#endif
	new = low;
	}
else if (new > high)
  	{
#ifdef VERBOSE
	printf("\t\tInterpolant (%.10f) too large: clamp to %.10f\n",new,high);
#endif
	new = high;
	}
#ifdef VERBOSE
else printf("\t\tInterpolant (%.10f) OK\n",new);
#endif
*res = new;
return(1);
}


double find_min_along(f,x,g,s,obj,grad,new_grad)
double *f;
double (*obj) ();
int *(*grad)();
struct vector *x,*g,*s,*new_grad;
/*
	f	- objective function value
	obj	- objective function
	x	- current point
	g	- gradient vector at x
	s	- search vector at x
	grad	- gradient function
*/
{
double new_f,alpha;		/* new function evaluation */
struct vector new_x;
double a1,a2;		/* limits of search - initially 0 and inf*/
double f_prime;		/* slope scalar at x */
double delta,alpha_hat;
int done,f2_ok,succeed,bracket;
double f1,f1_prime,comp,reference,ref_f;
static double fprev = 0;

if ((f1_prime = dot_product(g,s)) >= 0)
	{
	printf("**** INVALID GRADIENT FOR MINIMIZATION - %.10f ****\n",f1_prime);
	return(0);
	}
reference = f1_prime;

a1 = 0;
a2 = (F_MIN - *f)/(RHO * f1_prime);

f1 = *f;
ref_f = f1;

/*
	Estimate initial alpha
*/
if (fprev == 0) delta = *f - F_MIN;
else delta = fprev - *f;
if (delta < EPSILON10) delta = EPSILON10;
alpha = - 2 * delta / f1_prime;
if (alpha > a2) alpha = a2;

/*
	alpha should be 1.00 whenever possible, because the Newton
	step will result in faster convergence, near the solution,
	for BFGS methods
*/

#ifdef QUASI_NEWTON
if (alpha > 1.00) alpha = 1.00;
#endif

#ifdef VERBOSE
printf("Begin Line Search:\n");
printf("  --| Reduce %.10f with gradient %.10f\n",f1,f1_prime);
printf("  --| %.10f < alpha =  %.10f < %.10f\n",a1,alpha,a2);
#endif

/*
Loop conditions
	0 - still looking
	1 - successful
	2 - error, search failed to meet constraints
*/

bracket = 0;
done = 0;

while (!done)
	{
/*
	Step 1 : evaluate function at x + alpha * s
*/
		extend(&new_x,x,s,alpha);
		new_f = (*obj)(&new_x);
#ifdef VERBOSE
	        printf("\tObjective value %.10f at alpha = %.10f\n",
						new_f,alpha);
#endif
		f2_ok = 0;

/*	step2: check actual delta_f against expected	(First Goldstein Criterion)  */

		delta = new_f - ref_f;
		comp = RHO * alpha * reference;
#ifdef MONOTONIC
		if ((delta <= comp) && (new_f < f1))
#else
		if (delta <= comp)
#endif
			{
			(*grad)(&new_x,new_grad);
			f_prime = dot_product(new_grad,s);
#ifdef VERBOSE
		        printf("\tReduction OK - new gradient = %.10f\n",f_prime);
#endif
			f2_ok = 1;
/*
	Check termination conditions (Wolfe Test)
*/
#ifdef WOLFE_TEST
			if (fabs(f_prime) <= SIGMA * (-reference)) done = 1;
			else if (f_prime > 0)
				{
#ifdef VERBOSE
	        printf("\tPositive Gradient rejected - try interpolation\n");
#endif
				succeed = interpolate(a1,f1,f1_prime,alpha,new_f,
					f_prime,f2_ok,&alpha_hat);
				if (succeed == 1)
					{
					a2 = alpha;
					alpha = alpha_hat;
					bracket = 1;
					}
				else done = 2;
				}
#else
			if (f_prime >= SIGMA * reference) done = 1;
#endif
			else
				{
/*
	Extrapolate to point within interval (alpha,a2)
*/
#ifdef VERBOSE
	        printf("\tGradient rejected - extrapolate\n");

#endif
				succeed = extrapolate(a1,f1,f1_prime,alpha,
					new_f,f_prime,a2,&alpha_hat,bracket);
				if (succeed == 1)
					{
					a1 = alpha;
					f1_prime = f_prime;
					f1 = new_f;
					alpha = alpha_hat;
					}
				else done = 2;
				}
			}
		else
			{
/*
	Interpolate to new value within interval (a1,alpha)
*/
#ifdef VERBOSE
		        printf("\tObjective rejected - try interpolation\n");
#endif
			succeed = interpolate(a1,f1,f1_prime,alpha,new_f,
				f_prime,f2_ok,&alpha_hat);
			if (succeed == 1)
				{
				a2 = alpha;
				alpha = alpha_hat;
				bracket = 1;
				}
			else done = 2;
			}
	}
if (done == 1)
	{
#ifdef VERBOSE
	        printf("\tGradient accepted\nLine Search Successful\n");
#endif
	fprev = *f;
	*f = new_f;
	return(alpha);
	}
else 
	{
#ifdef VERBOSE
	        printf("Line Search Failed\n");
#endif
	return(0);
	}
}
